<?php

namespace Monarch;

/**
 * RGB color class.
 */
class RGB extends AccessibleColor {

  /**
   * The red component.
   *
   * @var float
   */
  protected $red = 0.0;

  /**
   * The green component.
   *
   * @var float
   */
  protected $green = 0.0;

  /**
   * The blue component.
   *
   * @var float
   */
  protected $blue = 0.0;

  /**
   * Factory function.
   *
   * @param float $red
   *   The red component.
   * @param float $green
   *   The green component.
   * @param float $blue
   *   The blue component.
   *
   * @return RGB
   *   The created RGB object.
   */
  public static function create(float $red, float $green, float $blue) : RGB {
    return new static($red, $green, $blue);
  }

  /**
   * {@inheritdoc}
   */
  public static function fromHex(string $hex) : RGB {
    $hex = strtolower($hex);

    if (preg_match('/#*([a-z0-9]{6,6}|[a-z0-9]{3,3})/', $hex, $matches)) {
      $c = [];

      $is_compact = strlen($matches[1]) === 3;

      preg_match_all($is_compact ? '/./' : '/../', $matches[1], $c);

      $c = $c[0];

      if ($is_compact) {
        foreach ($c as $j => $hex) {
          $c[$j] = $hex . $hex;
        }
      }

      foreach ($c as $j => $hex) {
        $c[$j] = ((float) hexdec($hex));
      }

      return static::create($c[0], $c[1], $c[2]);
    }

    throw new \Exception('Invalid hex format!');
  }

  /**
   * Constructor function.
   *
   * @param float $red
   *   The red component.
   * @param float $green
   *   The green component.
   * @param float $blue
   *   The blue component.
   */
  public function __construct(float $red, float $green, float $blue) {
    $this->red = $red;
    $this->green = $green;
    $this->blue = $blue;
  }

  /**
   * Clamp and return the red value.
   *
   * @return float
   *   The clamped value.
   */
  public function r() : float {
    return max(0.0, min(255.0, $this->red));
  }

  /**
   * Clamp and return the green value.
   *
   * @return float
   *   The clamped value.
   */
  public function g() : float {
    return max(0.0, min(255.0, $this->green));
  }

  /**
   * Clamp and return the blue value.
   *
   * @return float
   *   The clamped value.
   */
  public function b() : float {
    return max(0.0, min(255.0, $this->blue));
  }

  /**
   * {@inheritdoc}
   */
  public function toRgb() : RGB {
    return new RGB($this->red, $this->green, $this->blue);
  }

  /**
   * {@inheritdoc}
   */
  public function toHsl() : HSL {
    $r = $this->r() / 255.0;
    $g = $this->g() / 255.0;
    $b = $this->b() / 255.0;

    $cmax = max($r, $g, $b);
    $cmin = min($r, $g, $b);
    $delta = $cmax - $cmin;

    if (!$delta) {
      $hue = 0.0;
    }
    elseif ($cmax === $r) {
      $hue = 60.0 * fmod(($g - $b) / $delta, 6.0);
    }
    elseif ($cmax === $g) {
      $hue = 60.0 * (($b - $r) / $delta + 2.0);
    }
    elseif ($cmax === $b) {
      $hue = 60.0 * (($r - $g) / $delta + 4.0);
    }
    else {
      $hue = 0.0;
    }

    $lightness = ($cmax + $cmin) / 2;

    if ($delta) {
      $saturation = $delta / (1.0 - abs(2.0 * $lightness - 1.0));
    }
    else {
      $saturation = 0.0;
    }
    return new HSL($hue, $saturation, $lightness);
  }

  /**
   * Convert to string.
   *
   * @return string
   *   The string representation of the class.
   */
  public function __toString() : string {
    $c = [];

    foreach ([$this->r(), $this->g(), $this->b()] as $i => $v) {
      $c[$i] = str_pad(dechex((int) $v), 2, '0', STR_PAD_LEFT);
    }

    return '#' . implode('', $c);
  }

  /**
   * {@inheritdoc}
   */
  public function toHex() : string {
    return (string) $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getLuminance() : float {
    $r = $this->r() / 255.0;
    $g = $this->g() / 255.0;
    $b = $this->b() / 255.0;

    return 0.2126 * ($r < 0.03928 ? ($r / 12.92) : pow(($r + 0.055) / 1.055, 2.4)) +
      0.7152 * ($g < 0.03928 ? ($g / 12.92) : pow(($g + 0.055) / 1.055, 2.4)) +
      0.0722 * ($b < 0.03928 ? ($b / 12.92) : pow(($b + 0.055) / 1.055, 2.4));
  }

  /**
   * {@inheritdoc}
   */
  public function getLightness() : float {
    return $this->toHsl()->l();
  }

  /**
   * {@inheritdoc}
   */
  public function setLightness(float $lightness) : RGB {
    $tmp = $this->toHsl()->setLightness($lightness)->toRgb();

    $this->red = $tmp->red;
    $this->green = $tmp->green;
    $this->blue = $tmp->blue;

    return $this;
  }

}
